# Copyright (c) 2010 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from webob import Response, Request
from webob.exc import HTTPBadRequest, HTTPForbidden, HTTPNotFound, \
    HTTPUnauthorized, HTTPNoContent, HTTPAccepted, HTTPCreated
from swift.common.utils import get_logger, split_path, hash_path, get_param, \
    TRUE_VALUES
from urllib import quote
try:
    import simplejson as json
except ImportError:
    import json


MIN_TTL = 900
MAX_TTL = 3155692600



class Origin(object):
    '''
    Origin server for public containers
    '''

    def __init__(self, app, conf):
        self.app = app
        self.conf = conf
        self.logger = get_logger(conf, log_route='origin')

        self.origin_db_hosts = [host for host in
            conf.get('origin_db_hosts', '').split(',') if host]
        self.origin_cdn_hosts = [host for host in
            conf.get('origin_cdn_hosts', '').split(',') if host]
        self.origin_account = conf.get('origin_account')
        self.admin_key = conf.get('origin_admin_key')
        self.origin_prefix = conf.get('origin_prefix', '/origin/')
        self.cdn_hostname = conf.get('cdn_uri', 'cf1.rackcdn.com')

    def _valid_setup(self):
        valid_setup =  bool(self.origin_account and
                            self.origin_db_hosts and
                            self.origin_cdn_hosts)
        if not valid_setup:
            try:
                self.logger.critical( _('Invalid origin conf file!'))
            except Exception:
                pass
        return valid_setup

    def __call__(self, env, start_response):
        '''
        Accepts a standard WSGI application call.
        There are 2 types of requests that this middleware will affect.
        1. Requests to CDN 'database' that will enable, list, etc. containers
           for the CDN.
        2. Requests (GETs, HEADs) from CDN provider to publicly available
           containers.
        The types of requests can be determined by looking at the hostname of
        the incoming call.
        Wraps env in webob.Request object and passes it down.

        :param env: WSGI environment dictionary
        :param start_response: WSGI callable
        '''
        if not self._valid_setup():
            return self.app(env, start_response)
        req = Request(env)
        host = req.host.split(':')[0]

        if host in self.origin_db_hosts:
            return self.handle_origin_db_request(env, req)(env, start_response)
        if host in self.origin_cdn_hosts:
            return self.handle_origin_cdn_request(req)(env, start_response)

        if env.get('PATH_INFO', '').startswith(self.origin_prefix):
            return self.handle_origin_admin(req)(env, start_response)
        return self.app(env, start_response)

    def _getHashFromCdn(self, req):
        pass

    def getContainers(self, req):
        pass

    def validatePutRequest(self):
        pass
    def parse_headers(self, env):
        '''
        Builds json string for the object that will be stored in
        /.cdn/.hash/<hash>
        '''
        cdn_obj = {'ttl': env.get('X-TTL', ''),
                   'logs': env.get('X-CDN-Enabled'),
                   'enabled': cdn_enabled}
        return json.dumps(cdn_obj)

    def is_origin_admin(self, req):
        """
        Returns True if the admin specified in the request represents the
        .origin_admin.

        :param req: The webob.Request to check.
        :param returns: True if .origin_admin.
        """
        return self.admin_key and \
           req.headers.get('x-origin-admin-user') == '.origin_admin' and \
           req.headers.get('x-origin-admin-key') == self.admin_key

    def handle_origin_admin(self, req):
        """
        Handles the POST /origin/.prep call for preparing the backing store
        Swift cluster for use with the origin subsystem. Can only be called by
        .origin_admin

        :param req: The webob.Request to process.
        :returns: webob.Response, 204 on success
        """
        if not self.is_origin_admin(req):
            return HTTPForbidden(request=req)
        try:
            version, account, container = split_path(req.path, 1, 3, True)
        except ValueError:
            return HTTPNotFound(request=req)
        if account == '.prep':
            path = quote('/v1/%s' % self.origin_account)
            resp = self.make_request(req.environ, 'PUT',
                                     path).get_response(self.app)
            if resp.status_int // 100 != 2:
                raise Exception('Could not create the main origin account: %s %s' %
                                (path, resp.status))
            #TODO : is it worth it making 0-15 containers so they don't get
            # too big?  for now i'm gona leave it with just one.
            for cont_name in ['.id', '.hash']:
                path = quote('/v1/%s/%s' % (self.origin_account, cont_name))
                resp = self.make_request(req.environ, 'PUT',
                                         path).get_response(self.app)
                if resp.status_int // 100 != 2:
                    raise Exception('Could not create %s container: %s %s' %
                                    (cont_name, path, resp.status))
            return HTTPNoContent(request=req)
        return HTTPNotFound(request=req)

    def _gen_listing_content_type(self, cdn_enabled, ttl, logs_enabled, old_cdn_id=None):
        if old_cdn_id:
            return 'x-cdn-0/%(cdn_enabled)s-%(ttl)d-%(log_ret)s-%(old_id)s' % {
                'cdn_enabled': cdn_enabled, 'ttl': ttl,
                'log_ret': logs_enabled, 'old_id': old_cdn_id}
        return 'x-cdn-1/%(cdn_enabled)s-%(ttl)d-%(log_ret)s' % {
            'cdn_enabled': cdn_enabled, 'ttl': ttl, 'log_ret': logs_enabled}

    def _get_listing_content_type(self, account, container, data, output_format,
                                  only_cdn_enabled=False):
        '''
        Given a munged up content-type will return a str in desired format
        :raises: InvalidContentType
        '''
        if data.startswith('x-cdn-0/'):
            try:
                cdn_enabled, ttl, log_ret, old_id = \
                    data[len('x-cdn-0/'):].split('-')
            except ValueError:
                raise InvalidContentType('Invalid Content-Type: %s/%s: %s' %
                    (account, container, data))
            cdn_uri = 'http://c%s.r%d.%s' % (old_id, int(old_id)%100,
                                             self.cdn_hostname)
            cdn_uri_https = 'https://c%s.ssl.%s' % (old_id, self.cdn_hostname)
        elif data.startswith('x-cdn-1/'):
            try:
                cdn_enabled, ttl, log_ret = data[len('x-cdn-1/'):].split('-')
            except ValueError:
                raise InvalidContentType('Invalid Content-Type: %s/%s: %s' %
                    (account, container, data))
            hsh = hash_path(account, container)
            cdn_uri = 'http://h%s.r%d.%s' % (hsh, int(hsh[-2:],16) % 100,
                                             self.cdn_hostname)
            cdn_uri_https = 'https://h%s.ssl.%s' % (hsh, self.cdn_hostname)
        else:
            raise InvalidContentType('Invalid Content-Type: %s/%s: %s' %
                                     (account, container, data))

        if only_cdn_enabled and not cdn_enabled:
            return None
        if output_format == 'json':
            pass
# DAVID HERE YOU ARE
            

    def origin_db_get(self, env, req):
        '''
        Handles GETs to the Origin database
        The only part of the path this pays attention to is the account.
        TODO: BTW- existing origin returns 400 if you send marker and no limit
        '''
        try:
            account = req.path.split('/')[2]
        except IndexError:
            return HTTPBadRequest('Invalid request. '
                                  'URI format: /<api version>/<account>')
        #TODO: make sure to test with unicode container names
        marker = quote(get_param(req, 'marker', default=''))
        list_format = get_param(req, 'format')
        enabled_only = get_param(req, 'enabled', default='false') in TRUE_VALUES
        limit = get_param(req, 'limit')
        if limit:
            try:
                limit= int(limit)
            except ValueError:
                return HTTPBadRequest('Invalid limit, must be an integer')
        listing_path = '/v1/%s/%s?format=json&marker=%s' % \
                       (self.origin_account, account, marker)
        self.logger.info("the path: %s" % listing_path)
        # not putting limit in request because may have to filter on cdn_enabled 
        resp = self.make_request(env, 'GET',
                                 listing_path).get_response(self.app)
        self.logger.info("the resp: %s" % resp.body)
        response_body = ''
        if resp.status_int // 100 = 2:
            cont_listing = json.loads(resp.body)
            # TODO: is it ok to load the whole thing? do i have a choice?
            listing_formatted = []
            for listing_dict in cont_listing:
                formatted_data = self.format_listing(listing_dict, list_format, enabled_only)
                if formatted_data is None:
                    continue
                if limit and len(listing_formatted) < limit:
                    listing_formatted.append(formatted_data)
                else:
                    break
            self.logger.info("the output: %s" % listing_formatted)
        else:
            return HTTPNotFound(request=req)
        return Response(request=req)

    def origin_db_puts_posts(self, env, req):
        '''
        Handles PUTs and POSTs of new containers into Origin database
        '''
        #TODO: should I put in etags to verify that if got put correctly?
        # if so- what do I do about content-type stuff? but the listings aren't
        # that big of a deal
        try:
            version, account, container = split_path(req.path, 1, 3, True)
        except ValueError:
            return HTTPNotFound()
        ttl = '900'
        cdn_enabled = 'true'
        logs_enabled = 'false'
        #TODO: should I use a different hash suffix since these are public?
        hsh = hash_path(account, container)
        cdn_obj_path = quote('/v1/%s/.hash/%s' % (self.origin_account, hsh))
        # get defaults
        resp = self.make_request(env, 'GET',
            cdn_obj_path).get_response(self.app)
        if resp.status_int // 100 == 2:
            try:
                current_data = json.loads(resp.body)
                ttl = current_data.get('ttl', ttl)
                cdn_enabled = current_data.get('cdn_enabled', cdn_enabled)
                logs_enabled = current_data.get('logs_enabled',
                                                logs_enabled)
            except ValueError:
                pass # TODO: ignore json errors in the data files, ok right?
        elif req.method == 'POST':
            return HTTPNotFound(request=req)
        try:
            ttl = int(req.headers.get('X-TTL', ttl))
        except ValueError:
            return HttpBadRequest(_('Invalid X-TTL, must be integer'))
        if ttl < MIN_TTL or ttl > MAX_TTL:
            # TODO: this isn't exactly whats currently there. It only errors on
            # invalid TTLs if the enabled is true or being set to true
            return HTTPForbidden(
                _('Invalid X-TTL, must be between %(min)s and %(max)s'),
                {'min': MIN_TTL, 'max': MAX_TTL})
        logs_enabled = req.headers.get('X-Log-Retention',
            logs_enabled).lower() in TRUE_VALUES and 'true' or 'false'
        cdn_enabled = req.headers.get('X-CDN-Enabled',
            cdn_enabled).lower() in TRUE_VALUES and 'true' or 'false'

        cdn_data = {'account': account, 'container': container, 'ttl': ttl,
                    'logs_enabled': logs_enabled, 'cdn_enabled': cdn_enabled}
        cdn_obj_data = json.dumps(cdn_data)
        # this is always a PUT because a POST needs to update the contents of the file
        cdn_obj_resp = self.make_request(env, 'PUT', cdn_obj_path,
            body=cdn_obj_data).get_response(self.app)
        self.logger.info('hash req: %s : %s : %s' % (cdn_obj_path,
                                                     cdn_obj_data,
                                                     cdn_obj_resp))

        if cdn_obj_resp.status_int // 100 != 2:
            return cdn_obj_resp #TODO: this is not right, need to build a better response

        listing_cont_path = quote('/v1/%s/%s' % (self.origin_account, account))
        resp = self.make_request(req.environ, 'HEAD',
                                 listing_cont_path).get_response(self.app)
        if resp.status_int == 404:
            resp = self.make_request(req.environ, 'PUT',
                                     listing_cont_path).get_response(self.app)
            if resp.status_int // 100 != 2:
                raise Exception('Could not create account within '
                    'origin db: %s %s' % (listing_cont_path, resp.status))

        cdn_list_path = quote('/v1/%s/%s/%s' % (self.origin_account,
                                                account, container))

        cdn_list_resp = self.make_request(env, req.method, cdn_list_path,
            headers = {'Content-Type':
                self._gen_listing_content_type(cdn_enabled, ttl, logs_enabled),
                'Content-Length': 0}).get_response(self.app)

        if cdn_list_resp.status_int // 100 != 2:
            return cdn_list_resp #TODO: this is not right need to build a better response

        headers = {'X-CDN-URI': 'http://h%s.r%d.%s' % (hsh,
                                                       int(hsh[-2:],16) % 100,
                                                       #TODO: are these always 2 digits?
                                                       self.cdn_hostname),
                   'X-CDN-SSL-URI': 'https://h%s.%s' % (hsh, self.cdn_hostname)}

        if req.method == 'POST':
            return HTTPAccepted(request=req, headers=headers)
        else:
            return HTTPCreated(request=req, headers=headers)


    def _get_content_type_str(self, cdn_enabled, ttl, log_retention):
        '''
        Returns a string that will be saved as content-type in .cdn container
        '''
        return 'x-cdn/%(cdn_enabled)s-%(ttl)d-%(log_ret)s' % {
            'cdn_enabled': cdn_enabled, 'ttl': ttl, 'log_ret': log_retention}


    def handle_origin_db_request(self, env, req):
        '''
        This handles requests from a user to activate cdn access for their
        containers, list them, etc.
        '''
        if 'swift.authorize' in req.environ:
            aresp = req.environ['swift.authorize'](req)
            if aresp:
                return aresp

        if req.method in  ('PUT', 'POST'):
            return self.origin_db_puts_posts(env, req)
        if req.method in  ('GET'):
            return self.origin_db_get(env, req)
        return self.app

    def handle_origin_cdn_request(self, req):

        return self.app

    def make_request(self, env, method, path, body=None, headers=None):
        """
        Makes a new webob.Request based on the current env but with the
        parameters specified. Note that this request will be preauthorized.

        :param env: Current WSGI environment dictionary
        :param method: HTTP method of new request
        :param path: HTTP path of new request
        :param body: HTTP body of new request; None by default
        :param headers: Extra HTTP headers of new request; None by default

        :returns: webob.Request object
        (Stolen from Swauth: https://github.com/gholt/swauth)
        """
        newenv = {'REQUEST_METHOD': method, 'HTTP_USER_AGENT': 'SwiftOrigin'}
        for name in ('swift.cache', 'HTTP_X_CF_TRANS_ID'):
            if name in env:
                newenv[name] = env[name]
        newenv['swift.authorize'] = lambda req: None
        if not headers:
            headers = {}
        if body:
            return Request.blank(path, environ=newenv, body=body,
                                 headers=headers)
        else:
            return Request.blank(path, environ=newenv, headers=headers)

def filter_factory(global_conf, **local_conf):
    """Returns a WSGI filter app for use with paste.deploy."""
    conf = global_conf.copy()
    conf.update(local_conf)

    def origin(app):
        return Origin(app, conf)
    return origin
